from .abc import AbstractInitializer, InitializingCamera, InitializedPointCloud
from .dataset import InitializedCameraDataset, TrainableCameraDataset, TrainableInitializedCameraDataset
from .dust3r import Dust3rInitializer, Mast3rInitializer
from .colmap import ColmapSparseInitializer, ColmapDenseInitializer
from .depth import DepthInitializerWrapper, DepthAnythingV2InitializerWrapper
from .combinations import Dust3rAlign2ColmapDenseInitializer
from .combinations import DepthAnythingV2ColmapDenseInitializer, DepthAnythingV2ColmapSparseInitializer
from .combinations import DepthAnythingV2Dust3rInitializer, DepthAnythingV2Mast3rInitializer
from .combinations import DepthAnythingV2Dust3rAlign2ColmapDenseInitializer



from .colmap import ColmapSparseInitializer, ColmapDenseInitializer
from .dust3r import Dust3rInitializer, Dust3rAlign2Initializer, Mast3rInitializer
from .depth import AutoScaleDepthAnythingV2InitializerWrapper


def Dust3rAlign2ColmapDenseInitializer(
        model_path: str = "checkpoints/DUSt3R_ViTLarge_BaseDecoder_512_dpt.pth",
        batch_size: int = 1,
        niter: int = 300,
        schedule: str = 'linear',
        lr: float = 0.01,
        focal_avg: bool = True,
        scene_scale: float = 1.0,
        resize: int = 512,
        *args, **kwargs):
    return Dust3rAlign2Initializer(
        ColmapDenseInitializer(*args, **kwargs),
        model_path=model_path,
        batch_size=batch_size,
        niter=niter,
        schedule=schedule,
        lr=lr,
        focal_avg=focal_avg,
        scene_scale=scene_scale,
        resize=resize,
    )


def DepthAnythingV2ColmapSparseInitializer(
        input_size=518,
        encoder='vitl',
        checkpoints_folder='checkpoints',
        device="cuda",
        *args, **kwargs):
    return AutoScaleDepthAnythingV2InitializerWrapper(
        ColmapSparseInitializer(*args, **kwargs),
        input_size=input_size,
        encoder=encoder,
        checkpoints_folder=checkpoints_folder,
        device=device,
    )


def DepthAnythingV2ColmapDenseInitializer(
        input_size=518,
        encoder='vitl',
        checkpoints_folder='checkpoints',
        device="cuda",
        *args, **kwargs):
    return AutoScaleDepthAnythingV2InitializerWrapper(
        ColmapDenseInitializer(*args, **kwargs),
        input_size=input_size,
        encoder=encoder,
        checkpoints_folder=checkpoints_folder,
        device=device,
    )


def DepthAnythingV2Dust3rInitializer(
        input_size=518,
        encoder='vitl',
        checkpoints_folder='checkpoints',
        device="cuda",
        *args, **kwargs):
    return AutoScaleDepthAnythingV2InitializerWrapper(
        Dust3rInitializer(*args, **kwargs),
        input_size=input_size,
        encoder=encoder,
        checkpoints_folder=checkpoints_folder,
        device=device,
    )


def DepthAnythingV2Mast3rInitializer(
        input_size=518,
        encoder='vitl',
        checkpoints_folder='checkpoints',
        device="cuda",
        *args, **kwargs):
    return AutoScaleDepthAnythingV2InitializerWrapper(
        Mast3rInitializer(*args, **kwargs),
        input_size=input_size,
        encoder=encoder,
        checkpoints_folder=checkpoints_folder,
        device=device,
    )


def DepthAnythingV2Dust3rAlign2ColmapDenseInitializer(
        input_size=518,
        encoder='vitl',
        checkpoints_folder='checkpoints',
        device="cuda",
        *args, **kwargs):
    return AutoScaleDepthAnythingV2InitializerWrapper(
        Dust3rAlign2ColmapDenseInitializer(*args, **kwargs),
        input_size=input_size,
        encoder=encoder,
        checkpoints_folder=checkpoints_folder,
        device=device,
    )



from .abc import DepthInitializerWrapper
from .depth_anything_v2 import DepthAnythingV2InitializerWrapper
from .autoscale import AutoScaleDepthInitializerWrapper
from .combinations import AutoScaleDepthAnythingV2InitializerWrapper



import abc
import os
from typing import List, Tuple

import cv2
import numpy as np
import tifffile
import torch
import tqdm

from instantsplat.initializer import AbstractInitializer, InitializedPointCloud, InitializingCamera


class DepthInitializerWrapper(AbstractInitializer):
    def __init__(self, base_initializer: AbstractInitializer):
        self.base_initializer = base_initializer

    def to(self, device: torch.device) -> 'AbstractInitializer':
        self.base_initializer = self.base_initializer.to(device)
        return self

    @abc.abstractmethod
    def compute_depths(self, pointcloud: InitializedPointCloud, cameras: List[InitializingCamera]) -> List[torch.Tensor]:
        """Compute depth for the given point cloud and cameras."""
        raise NotImplementedError("Subclasses should implement this method.")

    def depth_path(self, image_path: str) -> str:
        return os.path.join(
            os.path.dirname(os.path.dirname(image_path)),
            "depths",
            os.path.splitext(os.path.basename(image_path))[0]+'.tiff'
        )

    def __call__(self, image_path_list: List[str]) -> Tuple[InitializedPointCloud, List[InitializingCamera]]:
        pointcloud, cameras = self.base_initializer.__call__(image_path_list)
        depths = self.compute_depths(pointcloud, cameras)
        cameras_with_depth = []
        for i, camera in enumerate(tqdm.tqdm(cameras, desc="Saving Depths")):
            depth_path = self.depth_path(camera.image_path)
            os.makedirs(os.path.dirname(depth_path), exist_ok=True)
            depth = depths[i]
            if depth is None:
                continue
            depth = depth.cpu().numpy()
            if depth_path.endswith('.tiff'):
                tifffile.imwrite(depth_path, depth)
                depth_scaled = np.repeat(((depth - depth.min()) / (depth.max() - depth.min()) * 255.0).astype(np.uint8)[..., np.newaxis], 3, axis=-1)
                cv2.imwrite(os.path.splitext(depth_path)[0] + '.png', depth_scaled)
            else:
                cv2.imwrite(depth_path, depth)
            cameras_with_depth.append(camera._replace(depth_path=depth_path))
        return pointcloud, cameras_with_depth



from typing import List

import torch

from instantsplat.initializer import InitializedPointCloud, InitializingCamera
from .abc import DepthInitializerWrapper
from .utils import fov2focal, count, get_min_depth


class AutoScaleDepthInitializerWrapper(DepthInitializerWrapper):
    def __init__(self, base_initializer_wrapper: DepthInitializerWrapper):
        super().__init__(base_initializer_wrapper.base_initializer)
        self.base_initializer_wrapper = base_initializer_wrapper

    def autoscale_depth(self, raw_invdepth: torch.Tensor, pointcloud: InitializedPointCloud, camera: InitializingCamera) -> torch.Tensor:
        xyz = pointcloud.points
        fx, fy = fov2focal(camera.FoVx, camera.image_width), fov2focal(camera.FoVy, camera.image_height)
        cx, cy = camera.image_width / 2, camera.image_height / 2
        K = torch.tensor([[fx, 0, cx], [0, fy, cy], [0, 0, 1]], device=raw_invdepth.device)
        R = camera.R.to(dtype=xyz.dtype)
        T = camera.T.to(dtype=xyz.dtype)
        uvd = (K @ ((R @ xyz.T).T + T.unsqueeze(0)).T).T
        uv, d = (uvd[:, 0:2] / uvd[:, 2:3]).round().long(), uvd[:, 2]
        valid_mask = (0 <= uv[:, 0:2]).all(-1) & (uv[:, 0] < camera.image_width) & (uv[:, 1] < camera.image_height)
        uv, d = uv[valid_mask, :], d[valid_mask]
        d_count = count(uv, camera.image_height, camera.image_width)
        d_min = get_min_depth(uv, d, camera.image_height, camera.image_width)
        d_idx = d_count > 0 & (raw_invdepth > 1e-6) & (d_min > 1e-6)
        invd_raw = raw_invdepth[d_idx]
        invd_target = 1 / d_min[d_idx]
        center_raw = invd_raw.median()
        center_target = invd_target.median()
        extent_raw = (invd_raw.max() - center_raw).mean()
        extent_target = (invd_target.max() - center_target).mean()
        scale = extent_target / extent_raw
        shift = center_target - center_raw * scale
        return raw_invdepth * scale + shift

    def compute_depths(self, pointcloud: InitializedPointCloud, cameras: List[InitializingCamera]) -> List[torch.Tensor]:
        """Compute depth and auto scale according to the point cloud and cameras."""
        raw_depths = self.base_initializer_wrapper.compute_depths(pointcloud, cameras)
        return [self.autoscale_depth(raw_depth, pointcloud, camera) for raw_depth, camera in zip(raw_depths, cameras)]
